def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        temp_output_dir=True,
        input_files=['input.txt'],
        output_files=['output_hello.txt'])
