def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        input_files='input.txt',
        temp_output_dir=True,
        output_files=['client_output.txt'])
